# -*- coding: utf-8 -*-
"""
bpycv: computer vision utils for Blender
    https://github.com/DIYer22/bpycv
"""

from .__info__ import __version__

from .utils import ipython, objects_to_locals
from .hdri_manager import HdriManager
from .node_graph import activate_node_tree, Node
from .exr_image_parser import parser_exr

from .select_utils import bpy, scene, render, cam, world, get_objdf
from .statu_recover import undo
from .object_utils import (
    activate_obj,
    remove_obj,
    edit_mode,
    subdivide,
    duplicate,
    load_obj,
)
from .material_utils import set_inst_material, load_hdri_world, build_tex
from .render_utils import (
    set_cycles_compute_device_type,
    set_image_render,
    set_annotation_render,
    render_image,
    render_data,
)
from .scene_setting import (
    remove_useless_data,
    clear_all,
    set_shading_mode,
    add_stage,
    add_environment_box,
)
from .camera_utils import get_cams, set_cam_pose, set_cam_intrinsic, get_cam_intrinsic
from .pose_utils import (
    T_bcam2cv,
    homo_coord,
    set_matrix_world,
    get_pose_in_cam,
    set_pose_in_cam,
    get_6d_pose,
)
from .physic_utils import set_origin_and_record_old_v0
from .domain_randomzition import texture_bsdf_dr, load_distractor



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Sat Dec 28 20:51:37 2019
"""

from boxx import *

import bpy


def load_obj(filepath):
    """
    Auto load mesh file, support .stl, .obj,

    Parameters
    ----------
    filepath : str
        filepath.

    Returns
    -------
    obj

    """
    ext = filepath[filepath.rindex(".") + 1 :]
    ext_to_import_func = {
        "stl": bpy.ops.import_mesh.stl,
        "obj": bpy.ops.import_scene.obj,
        "dae": bpy.ops.wm.collada_import,
        "ply": bpy.ops.import_mesh.ply,
    }
    import_func = ext_to_import_func[ext]
    import_func(filepath=filepath)
    assert len(bpy.context.selected_objects) == 1, f'load "{filepath}" failed!'
    return bpy.context.selected_objects[-1]


def is_obj_valid(obj):
    try:
        dir(obj)
        return True
    except ReferenceError:
        return False


class activate_obj(object):
    def __init__(self, obj):
        self.current_obj = bpy.context.view_layer.objects.active
        bpy.context.view_layer.objects.active = obj

    def __enter__(self):
        return self

    def __exit__(self, typee, value, traceback):
        if is_obj_valid(self.current_obj):
            bpy.context.view_layer.objects.active = self.current_obj


def remove_obj(obj_or_str):
    if isinstance(obj_or_str, str):
        obj = bpy.data.objects[obj_or_str]
    else:
        obj = obj_or_str
    bpy.data.objects.remove(obj)


class edit_mode:
    def __init__(self, obj=None, mode="EDIT"):
        self.obj = obj
        self.mode = mode

    def __enter__(self):
        if self.obj is None:
            self.obj = bpy.context.object
        self.activate = activate_obj(self.obj)
        self.activate.__enter__()
        self.old_mode = self.obj.mode
        bpy.ops.object.mode_set(mode=self.mode)

    def __exit__(self, typee, value, traceback):
        bpy.ops.object.mode_set(mode=self.old_mode)
        self.activate.__exit__(typee, value, traceback)


def subdivide(obj, number_cuts=2):
    with edit_mode(obj):
        bpy.ops.mesh.subdivide(number_cuts=number_cuts)


def duplicate(obj, copy_data=False, collection=None):
    if collection is None:
        collection = bpy.context.collection
    new_obj = obj.copy()
    if copy_data:
        new_obj.data = obj.data.copy()
    collection.objects.link(new_obj)
    return new_obj


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Sun Jan 12 17:06:39 2020
"""
import boxx
from boxx import *
from boxx import os, setTimeout, mapmt, timegap, sleep, pathjoin

import random
import glob

from bs4 import BeautifulSoup
import requests
import urllib.parse as urlparse

bs = BeautifulSoup
rq = requests


def url2dict(url):
    parsed = urlparse.urlparse(url)
    return urlparse.parse_qs(parsed.query)


class HdriManager:
    exts = ["hdr", "exr"]

    def __init__(
        self,
        hdri_dir="./bpycv_hdri_cache",
        resolution="4k",
        category="all",
        download=False,
        debug=False,
    ):
        """
        Download and manage hdri file from https://hdrihaven.com/hdris/

        Parameters
        ----------
        hdri_dir : str, optional
            hdri dir. The default is "./bpycv_hdri_cache".
        resolution : str, optional
            choice [1k, 2k, 4k, 8k, 16k, 19k]. The default is "4k".
        category : str, optional
            refer to https://hdrihaven.com/hdris/ side bar
            choice one [All, Outdoor, Architecture, Building, Europe, 
                        Field, Forest, Grass, Hill, Park, Path, River, 
                        Road, Rock, Sun1, Tree, View, Skies, Indoor, 
                        Studio, Nature, Urban, Night, Sunrise/Sunset, 
                        Morning/Afternoon, Midday, Clear, Partly Cloudy, 
                        Overcast, High Contrast, Medium Contrast, 
                        Low Contrast, Natural Light, Artificial Light]. 
            The default is "all".
        download : bool, optional
            If True, auto download from https://hdrihaven.com/hdris/ 
            by another threading using requesets.
            The default is False.
        """
        self.resolution = resolution
        self.category = category.lower().replace("/", "-")
        self.hdri_dir = hdri_dir
        os.makedirs(hdri_dir, exist_ok=True)
        self.downloading = download
        self.debug = debug
        if self.downloading:
            print('Starting download ".hdr" file from "hdrihaven.com" in side threads')
            if debug:
                self.prepare()
            else:
                setTimeout(self.prepare)
        self.set_hdr_paths()

    def set_hdr_paths(self):
        self.all_paths = sorted(
            sum(
                [
                    glob.glob(os.path.join(self.hdri_dir, f"*.{ext}"))
                    for ext in self.exts
                ],
                [],
            )
        )
        if self.category == "all":
            self.hdr_paths = sorted(self.all_paths)
            return

        listt = []
        for path in self.all_paths:
            fname = boxx.filename(path)
            name = fname.split(".")[0]
            listt.append(
                dict(
                    name=name,
                    res=name.split("_")[-1],
                    cats=fname.split(".")[1].split("="),
                    tags=fname.split(".")[2].split("="),
                    path=path,
                )
            )
        self.df = boxx.pd.DataFrame(listt)
        if self.category == "all":
            hdr_paths = self.df.path
        else:
            hdr_paths = self.df[
                self.df.cats.apply(lambda cats: self.category in cats)
            ].path

        self.hdr_paths = sorted(hdr_paths)

    def __len__(self):
        if self.downloading:
            self.set_hdr_paths()
        return len(self.hdr_paths)

    def __getitem__(self, i):
        if self.downloading:
            self.set_hdr_paths()
        return self.hdr_paths[i]

    def sample(self):
        if self.downloading:
            self.set_hdr_paths()
        while not len(self.hdr_paths):
            assert (
                self.downloading
            ), f'No hdri file in "{self.hdri_dir}", make sure HdriManager(download=True)'
            self.set_hdr_paths()
            if timegap(5, 'waiting for download ".hdr" file'):
                print('Waiting for download first ".hdr" file....')
            sleep(0.1)
        return random.choice(self.hdr_paths)

    def prepare(self,):
        resolution = self.resolution
        category = self.category
        hdri_dir = self.hdri_dir
        url = f"https://hdrihaven.com/hdris/category/?c={category}"
        page = rq.get(url, timeout=5)
        html = BeautifulSoup(page.text, features="html.parser")
        hrefs = [a["href"] for a in html.find(id="item-grid").find_all("a")]

        names = [url2dict(href)["h"][0] for href in hrefs]

        #  'https://hdrihaven.com/files/hdris/tv_studio_4k.hdr'
        def download(name):
            try:
                if self.debug:
                    print(name)
                prefix = f"{name}_{resolution}"
                paths = boxx.glob(os.path.join(hdri_dir, prefix + "*"))
                if len(paths):
                    return paths[0]
                url = f"https://hdrihaven.com/hdri/?h={name}"
                html = BeautifulSoup(
                    rq.get(url, timeout=5).text, features="html.parser",
                )
                href = [
                    a["href"]
                    for a in html.find_all("a")
                    if f"_{resolution}." in a.get("href")
                ][0]
                cats = [
                    a.text
                    for a in html.find(text="Categories:").parent.parent.find_all("a")
                ]
                tags = [
                    a.text for a in html.find(text="Tags:").parent.parent.find_all("a")
                ]
                name = f"{prefix}.{'='.join(cats)}.{'='.join(tags)}.{href[-3:]}"

                path = pathjoin(hdri_dir, name)
                r = rq.get(href, timeout=5)
                assert r.status_code == 200
                os.makedirs(hdri_dir, exist_ok=True)
                with open(path, "wb") as f:
                    f.write(r.content)
                return path
            except Exception as e:
                if self.debug:
                    boxx.pred - name
                    boxx.g()
                raise e

        _names = names[:]
        random.shuffle(_names)
        if self.debug:
            list(map(download, _names,))
        else:
            mapmt(download, _names, pool=1)
        self.set_hdr_paths()
        self.downloading = False
        print("Download hdri threads has finished!")

    @classmethod
    def test(cls):
        hdri_dir = "/tmp/hdri"
        hm = HdriManager(
            hdri_dir=hdri_dir, category="indoor", download=False, debug=True
        )
        for i in range(10):
            hdri = hm.sample()
            print(hdri)
            assert hm.category in hdri
        boxx.g()


if __name__ == "__main__":
    hdri_dir = "/tmp/hdri"
    hm = HdriManager(hdri_dir=hdri_dir, download=True, debug=True)
    hdri = hm.sample()
    print(hdri)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Sat Dec 28 21:38:05 2019
"""

import boxx
from boxx import *
from boxx import listdir, pathjoin

import bpy
import random

from .utils import encode_inst_id
from .statu_recover import StatuRecover
from .node_graph import activate_node_tree, Node


class set_inst_material(StatuRecover):
    def __init__(self):
        StatuRecover.__init__(self)

        self.set_attr(bpy.data.worlds[0], "use_nodes", False)
        objs = [obj for obj in bpy.data.objects if obj.type in ("MESH", "CURVE")]
        for obj_idx, obj in enumerate(objs):
            inst_id = obj.get("inst_id", 0)  # default inst_id is 0
            color = tuple(encode_inst_id.id_to_rgb(inst_id)) + (1,)

            material_name = "auto.inst_material." + obj.name
            material = bpy.data.materials.new(material_name)
            material["is_auto"] = True
            material.use_nodes = True
            material.node_tree.nodes.clear()
            with activate_node_tree(material.node_tree):
                Node("ShaderNodeOutputMaterial").Surface = Node(
                    "ShaderNodeEmission", Color=color
                ).Emission

            self.replace_collection(obj.data.materials, [material])


def remove_mat(mat_or_str):
    if isinstance(mat_or_str, str):
        mat = bpy.data.objects[mat_or_str]
    else:
        mat = mat_or_str
    bpy.data.materials.remove(mat)


def load_hdri_world(hdri_path, random_rotate_z=False):
    """
    

    Parameters
    ----------
    hdri_path : str
        pass.
    random_rotate_z : bool, optional
        How deg of hdri rotation. The default is False.

    Returns
    -------
    env_node
    """
    world = bpy.data.worlds[0]
    world.use_nodes = True
    world.node_tree.nodes.clear()
    with activate_node_tree(world.node_tree):
        env_node = Node("ShaderNodeTexEnvironment")
        bg_node = Node("ShaderNodeBackground")
        output_node = Node("ShaderNodeOutputWorld")
        env_node.image = bpy.data.images.load(hdri_path)
        bg_node.Color = env_node.Color
        output_node.Surface = bg_node.Background
        if random_rotate_z:
            rotate_deg = random.random() * 720
            coord = Node("ShaderNodeTexCoord")
            mapping = Node("ShaderNodeMapping", vector_type="TEXTURE")
            mapping.Vector = coord.Object
            coord.location = -900, 0
            mapping.location = -600, 0
            env_node.Vector = mapping.Vector
            mapping.node.inputs["Rotation"].default_value.z = rotate_deg
    return env_node


def alias_texture_name_to_name(texture_name):
    texture_name = texture_name.lower()
    return texture_name


def get_texture_paths(texture_dir):
    bnames = listdir(texture_dir)
    for i, cs in enumerate(zip(*bnames)):
        if len(set(cs)) != 1:
            break
    for _i, cs in enumerate(zip(*[b[::-1] for b in bnames])):
        if len(set(cs)) != 1:
            break
    texture_paths = {
        alias_texture_name_to_name(bname[i:-_i]): pathjoin(texture_dir, bname)
        for bname in bnames
    }
    texture_paths["name"] = bnames[0][:i] + "." + bnames[0][-_i:]
    return texture_paths


def build_tex(texture_dir):
    texture_paths = get_texture_paths(texture_dir)
    material_name = texture_paths.get("name", "load_mat")
    material = bpy.data.materials.new(material_name)
    material.use_nodes = True
    material.node_tree.nodes.clear()
    material.cycles.displacement_method = "BOTH"
    with activate_node_tree(material.node_tree):
        bsdf = Node("ShaderNodeBsdfPrincipled")
        output = Node("ShaderNodeOutputMaterial")
        output.Surface = bsdf.BSDF
        bsdf.location = 600, 300
        output.location = 900, -200

        coord = Node("ShaderNodeTexCoord")
        mapping = Node("ShaderNodeMapping", vector_type="TEXTURE")
        mapping.Vector = coord.Object
        coord.location = -900, 0
        mapping.location = -600, 0

        diff = Node(
            "ShaderNodeTexImage",
            projection="BOX",
            image=bpy.data.images.load(texture_paths["diff"]),
        )
        diff.Vector = mapping.Vector
        if texture_paths.get("ao"):
            ao_img = bpy.data.images.load(texture_paths["ao"])
            ao_img.colorspace_settings.name = "Non-Color"
            ao = Node("ShaderNodeTexImage", projection="BOX", image=ao_img)

            ao.Vector = mapping.Vector
            mix = Node("ShaderNodeMixRGB", Fac=0.1)
            mix.Color1 = diff.Color
            mix.Color2 = ao.Color
            color_output = mix.Color
        else:
            color_output = diff.Color
        bsdf["Base Color"] = color_output

        rough_img = bpy.data.images.load(texture_paths["rough"])
        rough_img.colorspace_settings.name = "Non-Color"
        rough = Node("ShaderNodeTexImage", projection="BOX", image=rough_img)
        rough.Vector = mapping.Vector
        if texture_paths.get("rough_ao"):
            rough_ao_img = bpy.data.images.load(texture_paths["rough_ao"])
            rough_ao_img.colorspace_settings.name = "Non-Color"
            rough_ao = Node("ShaderNodeTexImage", projection="BOX", image=rough_ao_img)

            rough_ao.Vector = mapping.Vector
            rough_mix = Node("ShaderNodeMixRGB", Fac=0.5)
            rough_mix.Color1 = rough.Color
            rough_mix.Color2 = rough_ao.Color
            rough_output = rough_mix.Color
        else:
            rough_output = rough.Color

        bsdf["Roughness"] = rough_output

        nor_img = bpy.data.images.load(texture_paths["nor"])
        nor_img.colorspace_settings.name = "Non-Color"
        nor = Node("ShaderNodeTexImage", projection="BOX", image=nor_img)
        nor.Vector = mapping.Vector
        normal_map = Node("ShaderNodeNormalMap")
        normal_map.Color = nor.Color
        bsdf.Normal = normal_map.Normal

        disp_img = bpy.data.images.load(texture_paths["disp"])
        disp_img.colorspace_settings.name = "Non-Color"
        disp = Node("ShaderNodeTexImage", projection="BOX", image=disp_img)
        disp.Vector = mapping.Vector
        displacement = Node("ShaderNodeDisplacement", Midlevel=0, Scale=0)
        displacement.Height = disp.Color
        output.Displacement = displacement.Displacement
        mapping.node.inputs["Scale"].default_value = (1.0,) * 3
        mapping.node.inputs["Location"].default_value = [
            random.random() * 8 for _ in "xyz"
        ]

    return material


if __name__ == "__main__":
    pass



__version__ = "0.2.41"
__description__ = "Computer vision utils for Blender."
__license__ = "MIT"
__author__ = "DIYer22"
__author_email__ = "ylxx@live.com"
__maintainer__ = "DIYer22"
__maintainer_email__ = "ylxx@live.com"
__github_username__ = "DIYer22"
__url__ = "https://github.com/DIYer22/bpycv"
__support__ = "https://github.com/DIYer22/bpycv/issues"
__classifiers__ = [
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Multimedia :: Graphics :: 3D Rendering",
    "Topic :: Multimedia :: Graphics :: 3D Modeling",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Thu Dec 26 18:15:47 2019
"""
import boxx
from boxx import *
from boxx import greyToRgb, histEqualize, inpkg, np, pathjoin, savenp

with inpkg():
    from .pseudo_color import heatmap_to_pseudo_color
    from .utils import encode_inst_id

import os
import struct

import bpy
import cv2
import minexr
import scipy.io


class ExrReader(minexr.reader.MinExrReader):
    def _read_image(self):
        """
        Override original _read_image, since that one assumes ims are float16 but we use float32 here
        """
        H, C, W = self.shape

        dtype = self.channel_types[0]
        DS = np.dtype(dtype).itemsize
        SOFF = 8 + DS * W * C
        strides = (SOFF, DS * W, DS)
        nbytes = SOFF * H

        self.fp.seek(self.first_offset, 0)
        image = np.frombuffer(self.fp.read(nbytes), dtype=dtype, count=-1, offset=8)
        self.image = np.lib.stride_tricks.as_strided(image, (H, C, W), strides)

    def _read_header(self):
        """
        Override original _read_header, since that one doesn't allow for cycles' long attribute names
        """
        self.fp.seek(0)
        buf = minexr.buffer.BufferReader(self.fp.read(10000))

        # Magic and version and info bits
        magic, version, b2, b3, b4 = struct.unpack("<iB3B", buf.read(8))
        assert magic == 20000630, "Not an OpenEXR file."
        assert b3 == b4 == 0, "Not a single-part scan line file."
        assert b2 in (0, 4), "Not a single-part scan line file."

        # Header attributes
        self.attrs = self._read_header_attrs(buf)

        # Parse channels and datawindow
        self.compr = self._parse_compression(self.attrs)
        self.channel_names, self.channel_types = self._parse_channels(self.attrs)
        self.channel_map = {cn: i for i, cn in enumerate(self.channel_names)}
        H, W = self._parse_data_window(self.attrs)
        self.shape = (H, len(self.channel_names), W)
        self.first_offset = self._read_first_offset(buf)

        # Assert our assumptions
        assert self.compr == 0x00, "Compression not supported."
        assert len(set(self.channel_types)) <= 1, "All channel types must be equal."


class ExrImage:
    LIMIT_DEPTH = 6e4

    def __init__(self, exr_path):
        with open(exr_path, "rb") as fp:
            self.reader = ExrReader(fp)

    def get_rgb(self):
        return self.reader.select(["R", "G", "B"]).copy()

    def get_rgba(self):
        return self.reader.select(["R", "G", "B", "A"]).copy()

    def get_pseudo_color(self):
        depth = self.reader.select(["Z"]).copy().squeeze()
        limit_mask = depth < self.LIMIT_DEPTH
        depth = depth * limit_mask
        depth = depth / depth.max()
        depth[~limit_mask] = 1.1
        depth = 1 - depth
        return heatmap_to_pseudo_color(depth)

    def get_depth(self):
        # turn inf depth to 0
        depth = self.reader.select(["Z"]).copy().squeeze()
        limit_mask = depth < self.LIMIT_DEPTH
        depth = depth * limit_mask
        return depth

    def get_inst(self):
        rgb = self.get_rgb()
        inst = encode_inst_id.rgb_to_id(rgb)

        # if world.use_nodes is False, Blender will set background as a gray (0.05087609, 0.05087609, 0.05087609)
        gray_background_mask = (rgb[..., 0] != 0) & (rgb[..., 0] != 1)
        inst[gray_background_mask] = -1
        return inst


class ImageWithAnnotation(dict):
    def __init__(self, image=None, exr=None, **kv):
        super().__init__(**kv)
        self["image"] = image
        self["inst"] = exr.get_inst()
        self["depth"] = exr.get_depth()
        self["_raw_exr"] = exr

    def __getattribute__(self, key):
        if key in self:
            return self[key]
        return dict.__getattribute__(self, key)

    def vis(self):
        vis_list = []  # RGB float 0.~1.

        def vis_inst(inst):
            unique, _idxs = np.unique(inst, return_inverse=True)
            idxs = _idxs.reshape(inst.shape)
            return boxx.mapping_array(
                idxs, boxx.getDefaultColorList(len(unique), includeBackGround=True)
            )

        if "inst" in self:
            inst_vis = vis_inst(self["inst"])
            vis_list.append(inst_vis)

        if self.get("image") is not None:
            image = self["image"]
            vis_list.append(image[..., :3] / 255.0)

        depth_vis = self["_raw_exr"].get_pseudo_color()
        vis_list.append(depth_vis)
        vis = (np.concatenate(vis_list, 1) * 255).astype(np.uint8)
        return vis

    def save(self, dataset_dir="dataset", fname="0", save_blend=False):
        fname = str(fname)
        if self.get("inst") is not None:
            inst_dir = pathjoin(dataset_dir, "instance_map")
            os.makedirs(inst_dir, exist_ok=True)
            inst_path = pathjoin(inst_dir, fname + ".png")
            cv2.imwrite(inst_path, self["inst"].clip(0).astype(np.uint16))
        if self.get("depth") is not None:
            depth_dir = pathjoin(dataset_dir, "depth")
            os.makedirs(depth_dir, exist_ok=True)
            depth_path = pathjoin(depth_dir, fname)
            savenp(depth_path, self["depth"].astype(np.float16))
        if (
            self.get("image") is not None
            and self.get("inst") is not None
            and self.get("depth") is not None
        ):
            vis_dir = pathjoin(dataset_dir, "vis")
            os.makedirs(vis_dir, exist_ok=True)
            vis_path = pathjoin(vis_dir, fname + ".jpg")
            cv2.imwrite(vis_path, self.vis()[..., ::-1])
        if self.get("ycb_6d_pose") is not None:
            pose_dir = pathjoin(dataset_dir, "ycb_6d_pose")
            os.makedirs(pose_dir, exist_ok=True)
            pose_path = pathjoin(pose_dir, fname + ".mat")
            scipy.io.savemat(pose_path, self["ycb_6d_pose"])
        if save_blend:
            blend_dir = pathjoin(dataset_dir, "blend")
            os.makedirs(blend_dir, exist_ok=True)
            blend_path = pathjoin(blend_dir, fname + ".blend")
            bpy.ops.wm.save_mainfile(filepath=blend_path)
        # save image at last for unstable compute enviroment
        if self.get("image") is not None:
            image_dir = pathjoin(dataset_dir, "image")
            os.makedirs(image_dir, exist_ok=True)
            image_path = pathjoin(image_dir, fname + ".jpg")
            cv2.imwrite(image_path, self["image"][..., ::-1])


def parser_exr(exr_path):
    exr = ExrImage(exr_path)
    return exr


def test_parser_exr(exr_path="../tmp_exrs/cycles.exr"):
    return parser_exr(exr_path)


if __name__ == "__main__":
    from boxx import imread, show

    exr_path = "../example/tmp_exr/inst.exr"
    exr = parser_exr(exr_path)
    inst = exr.get_inst()
    png = np.uint8(boxx.mapping_array(inst % 20, boxx.getDefaultColorList(20)) * 255)

    ann = ImageWithAnnotation(png, exr)
    vis = ann.vis()
    show - vis



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Sat Dec 28 20:56:00 2019
"""

from boxx import *
from boxx import pd

import bpy

objs = bpy.data.objects
scene = bpy.data.scenes[0]
render = scene.render
cam = bpy.data.objects.get("Camera", None)
world = bpy.data.worlds[0]


def get_objdf():
    row_dics = []

    def obj_to_row(obj):
        return dict(key=obj.name, type=obj.type, obj=obj)

    for k, obj in bpy.data.objects.items():
        row_dic = obj_to_row(obj)
        row_dics.append(row_dic)

    return pd.DataFrame(row_dics)


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Sat Dec 28 21:33:28 2019
"""

import tempfile
from collections import OrderedDict

import bpy
from boxx import *
from boxx import imread, os, withattr

from .exr_image_parser import ImageWithAnnotation, parser_exr
from .material_utils import set_inst_material
from .pose_utils import get_6d_pose
from .statu_recover import StatuRecover, undo


def set_cycles_compute_device_type(compute_device_type="CUDA"):
    bpy.context.scene.cycles.device = "GPU"
    bpy.context.preferences.addons[
        "cycles"
    ].preferences.compute_device_type = compute_device_type
    bpy.context.preferences.addons["cycles"].preferences.get_devices()
    print(
        "compute_device_type =",
        bpy.context.preferences.addons["cycles"].preferences.compute_device_type,
    )
    for d in bpy.context.preferences.addons["cycles"].preferences.devices:
        d["use"] = True
        print(d["name"], d["use"])


class set_annotation_render(StatuRecover):
    def __init__(self):
        StatuRecover.__init__(self)
        # TODO detect whether in ssh X11 forward
        # if sysi.gui:  # mean "does the enviroment support GUI".
        # self.set_attr(render, "engine", "BLENDER_EEVEE")
        scene = bpy.data.scenes[0]
        render = scene.render
        if render.engine == "BLENDER_WORKBENCH":
            self.set_attr(render, "engine", "CYCLES")
        if render.engine == "BLENDER_EEVEE":
            # When enviroment not support GUI, Eevee will raise Exception("Unable to open a display")  (@Blender 2.81)
            self.set_attr(render, "engine", "BLENDER_EEVEE")
            self.set_attr(scene.eevee, "taa_render_samples", 1)
            self.set_attr(scene.eevee, "use_bloom", False)
        elif render.engine == "CYCLES":
            self.set_attr(render, "engine", "CYCLES")
            self.set_attr(scene.cycles, "samples", 1)
            self.set_attr(
                scene.view_layers["View Layer"].cycles, "use_denoising", False
            )
        self.set_attr(render, "film_transparent", True)
        self.set_attr(scene.render, "use_motion_blur", False)
        self.set_attr(scene.render, "tile_x", 256)
        self.set_attr(scene.render, "tile_y", 256)

        attrs = dict(
            file_format="OPEN_EXR",
            compression=0,
            color_mode="RGBA",
            color_depth="32",
            exr_codec="NONE",
            use_zbuffer=True,
        )
        self.set_attrs(render.image_settings, attrs)


class set_image_render(StatuRecover):
    def __init__(self):
        StatuRecover.__init__(self)
        scene = bpy.data.scenes[0]
        render = scene.render
        attrs = dict(file_format="PNG", compression=15)
        self.set_attrs(render.image_settings, attrs)


def render_image():
    render = bpy.data.scenes[0].render
    png_path = tempfile.NamedTemporaryFile().name + ".png"
    with set_image_render(), withattr(render, "filepath", png_path):
        print("Render image using:", render.engine)
        bpy.ops.render.render(write_still=True)
    image = imread(png_path)[..., :3]
    os.remove(png_path)
    return image


_render_image = render_image
befor_render_data_hooks = OrderedDict()

# @undo()
def render_data(render_image=True, render_annotation=True):
    scene = bpy.data.scenes[0]
    render = scene.render
    for hook_name, hook in befor_render_data_hooks.items():
        print(f"Run befor_render_data_hooks[{hook_name}]")
        hook()
    befor_render_data_hooks.clear()

    path = tempfile.NamedTemporaryFile().name
    render_result = {}
    if render_image:
        render_result["image"] = _render_image()
    if render_annotation:
        exr_path = path + ".exr"
        with set_inst_material(), set_annotation_render(), withattr(
            render, "filepath", exr_path
        ):
            print("Render annotation using:", render.engine)
            bpy.ops.render.render(write_still=True)
        render_result["exr"] = parser_exr(exr_path)
        os.remove(exr_path)
    result = ImageWithAnnotation(**render_result)
    if "render_6d_pose" and render_annotation:
        objs = [obj for obj in bpy.data.objects if "inst_id" in obj]
        ycb_6d_pose = get_6d_pose(objs, inst=result["inst"])
        result["ycb_6d_pose"] = ycb_6d_pose
    return result


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Mon Jan 13 18:39:35 2020
"""

from boxx import *


class activate_node_tree:
    activate_node_tree_stack = []

    def __init__(self, node_tree):
        self.activate_node_tree_stack.append(node_tree)
        self.node_tree = node_tree

    def __enter__(self):
        return self

    def __exit__(self, typee, value, traceback):
        assert self.activate_node_tree_stack.pop() is self.node_tree


def is_node_socket(obj):
    return type(obj).__name__.startswith("NodeSocket")


class Node(object):
    """
    Usage:
        
    >>> with activate_node_tree(material.node_tree):
            Node("ShaderNodeOutputMaterial").Surface = Node("ShaderNodeEmission", Color=color).Emission
    
    >>> with activate_node_tree(material.node_tree):
            Node(
                "ShaderNodeOutputMaterial",
                Surface=Node("ShaderNodeEmission", Color=color).Emission,
            )
    """

    def __init__(self, node, node_tree=None, **kv):
        assert len(
            activate_node_tree.activate_node_tree_stack
        ), "Node init must under `with activate_node_tree(node_tree):`"
        if node_tree is None:
            node_tree = activate_node_tree.activate_node_tree_stack[-1]
        if isinstance(node, str):
            node = node_tree.nodes.new(node)
            node["is_auto"] = True
        self.node = node
        self.node_tree = node_tree
        self.kv = kv
        self.set_input(kv)
        self.begin_to_set_by_attr = True

    def __getattr__(self, key):
        node = object.__getattribute__(self, "node")
        if key in node.outputs:
            return node.outputs[key]
        return object.__getattribute__(self, key)

    def __setattr__(self, key, value, *l):
        if self.__dict__.get("begin_to_set_by_attr"):
            self.set_kv(key, value)
        else:
            object.__setattr__(self, key, value, *l)

    def __getitem__(self, key):
        return self.node.outputs[key]

    def __setitem__(self, key, value):
        self.set_kv(key, value)

    def set_kv(self, k, v):
        if k in self.node.inputs:
            node_input = self.node.inputs[k]
            if is_node_socket(v):
                self.node_tree.links.new(v, node_input)
            else:
                node_input.default_value = v
        else:
            setattr(self.node, k, v)

    def set_input(self, dic=None, **kv):
        if dic is None:
            dic = {}
        kv.update(dic)
        for k, v in kv.items():
            self.set_kv(k, v)


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Sat Dec 28 20:39:59 2019
"""
import sys
import boxx
from boxx import np


class encode_inst_id:
    """
    Encode inst id represent as [0.~1.] float32 rgb, that blender can read as color to render output
    """

    max_depth = 20
    max_denominator = 2 ** max_depth

    @classmethod
    def id_to_rgb(cls, id):
        """
        Cover float[-2**20~2**20] to float32[0.~1.] rgb

        Parameters
        ----------
        id : int, float or numpy
            float[-2**20~2**20].

        Returns
        -------
        (3,)float32
            RGB for Blender.

        """
        if isinstance(id, (float, int)):
            id = np.array(id)
        rgb = np.zeros(id.shape + (3,), dtype=np.float64)
        rgb[..., 0] = id < 0

        absf = np.abs(id)
        int_part = absf // 1
        # rgb[..., 1] = 1 - 1 / (int_part + 1)
        poww = np.int32(np.log2(int_part + 1, dtype=np.float32)) + 1
        denominator = (2 ** poww).round()
        rgb[..., 1] = ((int_part - denominator // 2 + 1) * 2 + 1) / denominator

        rgb[..., 2] = absf % 1
        return rgb

    @classmethod
    def rgb_to_id(cls, rgb):
        is_negative = (-1) ** (rgb[..., 0] == 1)
        # int_part = (1 / (1 - rgb[..., 1]) - 1).round()

        bg_mask = rgb[..., 1] == 0
        rgb[bg_mask, 1] = 0.5

        numerator = (cls.max_denominator * rgb[..., 1]).round().astype(np.int32)
        low_bit = (numerator ^ (numerator - 1)) & numerator
        numerator_odd = numerator // low_bit
        idx_in_level = (numerator_odd - 1) / 2
        up = np.int32(np.log2(low_bit, dtype=np.float32))
        depth = cls.max_depth - up
        int_part = 2 ** (depth - 1) - 1 + idx_in_level
        int_part = np.int32(int_part)

        int_part = int_part * (~bg_mask)
        if rgb[..., 2].any():  # has float
            return is_negative * (int_part + rgb[..., 2])
        else:  # pure int
            return np.int32(is_negative * int_part)

    @classmethod
    def test(cls):
        inst0 = np.arange(-100, 210000)
        inst1 = np.random.randint(
            -cls.max_denominator / 2, cls.max_denominator / 2, (10000)
        )
        inst2 = np.float32(
            np.linspace(-cls.max_denominator / 2, cls.max_denominator / 2, 10000)
        )
        inst3 = np.float32(np.random.rand(10000) * 10000.3 - 1000)
        for inst in [inst0, inst1, inst2, inst3]:
            boxx.tree - inst
            for inst_id in inst.flatten():
                rgb = encode_inst_id.id_to_rgb(inst_id).astype(np.float32)
                # print(rgb, inst_id)
                recover = encode_inst_id.rgb_to_id(rgb)
                assert recover == inst_id, (inst_id, recover)
            boxx.mg()

        assert (
            encode_inst_id.rgb_to_id(np.float32([0, 0, 0])) == 0
        ), "Check background should decode to 0"


def ipython():
    import IPython

    IPython.embed()


def objects_to_locals(log=True):
    """
    Add variables which are objects name and ref objects to python console when debug.
    """
    import bpy

    frame = sys._getframe(1)
    local_vars = frame.f_locals
    new = {name.replace(".", "_"): obj for name, obj in bpy.data.objects.items()}
    local_vars.update(new)
    if log:
        print("New variables:")
        print("\t", ", ".join(new))


if __name__ == "__main__":
    from boxx import *

    encode_inst_id.test()



#!/usr/bin/env python3
import os
import sys
import argparse
from zcs import argument, ints
from zcs.config import CfgNode as CN


cfg = CN()
cfg.DIR = argument("/tmp/syn_dataset", str, "Generated dataset dir.")
cfg.IMG_NUM = argument(500, int, "How many number of images to generate?")
cfg.OBJ_NUM_DIST = argument(
    [2, 3, 3, 4], ints, "List to reveal distribution of objcts number per scene."
)
cfg.RESOLUTION = argument([1024, 1024], ints, "Height and width of image.")

cfg.DEBUG = False
cfg.MAX_INST = argument(1000, int, "Max instance number per category in one image.")


def get_default_cfg():
    return cfg.clone()


def get_arguments():
    parser = argparse.ArgumentParser(description="Build synthetic dataset by bpycv")
    parser.add_argument("--config", type=str, default=None, help="config file path")
    parser.add_argument(
        "opts",
        help="Modify config options using the command-line",
        default=None,
        nargs=argparse.REMAINDER,
    )
    argv_for_py = sys.argv[sys.argv.index("--") + 1 :] if "--" in sys.argv else []
    args = parser.parse_args(argv_for_py)
    return args


if __name__ == "__main__":
    from boxx import tree

    args = get_arguments()
    cfg = get_default_cfg()
    cfg.merge_from_list_or_str(args.opts)
    tree(cfg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Tue Feb 18 18:15:29 2020
"""

import boxx
from boxx import os
from boxx import timeit, pathjoin

import bpy
import time
import random
from abc import abstractmethod

uniform_by_mean = lambda mean, rate=0.2: random.uniform(-1, 1) * mean * rate + mean


class MetaDatasetGenerator:
    def __init__(self, cfg):
        self.cfg = cfg
        self._blender_setting()

    def _blender_setting(self):
        render = bpy.data.scenes[0].render
        render.engine = "CYCLES"
        render.resolution_x = self.cfg.RESOLUTION[1]
        render.resolution_y = self.cfg.RESOLUTION[0]

        bpy.context.scene.frame_end = 168
        bpy.context.scene.frame_set(0)

        if self.cfg.DEBUG:
            bpy.context.scene.view_layers[0].cycles.use_denoising = False
            bpy.context.scene.cycles.samples = 8
        bpy.context.view_layer.update()

    @abstractmethod
    def generate_one(self, dirr, index):
        pass

    def exist(self, dirr, index):
        return os.path.isfile(pathjoin(dirr, "image", f"{index}.jpg"))

    def resample_keys(self, keys):
        try:
            keys = __import__("brainpp_yl").split_keys_by_replica(keys[:])
        except ModuleNotFoundError:
            pass
        return keys

    def generate_all(self):
        abs_dir = os.path.abspath(self.cfg.DIR)
        keys = range(self.cfg.IMG_NUM)

        if self.cfg.DEBUG:
            self.generate_one(abs_dir, str(int(time.time())))
            boxx.pred(abs_dir)
            return
        keys = self.resample_keys(keys)
        for index in keys:
            if not self.exist(abs_dir, index):
                with timeit("generate_one"):
                    self.generate_one(abs_dir, index)


if __name__ == "__main__":
    pass



#!/usr/bin/env python3

import bpy
import mathutils

from .object_utils import activate_obj

SET_ORIGIN_HOOK_NAME = "set_back_origin"
OLD_V0_KEY = "old_object.data.vertices[0].co"


def set_origin_and_record_old_v0(obj, type="ORIGIN_CENTER_OF_VOLUME", center="MEDIAN"):
    obj[OLD_V0_KEY] = obj.data.vertices[0].co.copy()
    with activate_obj(obj):
        bpy.ops.object.origin_set(type=type, center=center)


def set_origin_by_vector(obj, new_origin_vector):
    return set_origin_by_point(
        obj, obj.matrix_world.to_translation() + new_origin_vector
    )


def set_origin_by_point(obj, point):
    """
    Not work when bpy.context.scene.frame_curren != 1
    """
    cursor_mat = bpy.context.scene.cursor.matrix.copy()
    bpy.context.scene.cursor.rotation_euler = 0, 0, 0
    bpy.context.scene.cursor.location = point
    with activate_obj(obj):
        bpy.ops.object.origin_set(type="ORIGIN_CURSOR", center="MEDIAN")
    bpy.context.scene.cursor.matrix = cursor_mat
    return obj


if __name__ == "__main__":
    pass



#!/usr/bin/env python3

from boxx import *
from boxx import deg2rad, np, pi

import bpy
import random
import mathutils


def get_cams():
    return [obj for obj in bpy.data.objects if obj.type == "CAMERA"]


def set_cam_pose(cam_radius=1, cam_deg=45, cam_x_deg=None, cam=None):
    """
    Set the camera pose according to the shape of hemisphere.

    Parameters
    ----------
    cam_radius : float, optional
        The distance of the camera from the origin 0,0,0. The default is 1m.
    cam_deg : float, optional
        The angle between the optical center and the XY-plane. The default is 45°.
    cam_x_deg : float, optional
        The angle between the optical center and the X-axis. The default is None.
    cam : Camera object, optional
        The default is first camera object.

    Returns
    -------
    cam : Camera object
    """
    cam_rad = deg2rad(cam_deg)
    if cam_x_deg is None:
        cam_x_deg = random.uniform(0, 360)
    cam_x_rad = deg2rad(cam_x_deg)
    z = cam_radius * np.sin(cam_rad)
    xy = (cam_radius ** 2 - z ** 2) ** 0.5
    x = xy * np.cos(cam_x_rad)
    y = xy * np.sin(cam_x_rad)
    cam = cam or get_cams()[0]
    cam.location = x, y, z
    cam.rotation_euler = pi / 2 - cam_rad, 0.1, pi / 2 + cam_x_rad
    cam.scale = (0.1,) * 3
    return cam


def get_cam_intrinsic(cam=None):
    """
    Refrence:
        https://blender.stackexchange.com/a/120063/86396
    """
    # BKE_camera_sensor_size
    def get_sensor_size(sensor_fit, sensor_x, sensor_y):
        if sensor_fit == "VERTICAL":
            return sensor_y
        return sensor_x

    # BKE_camera_sensor_fit
    def get_sensor_fit(sensor_fit, size_x, size_y):
        if sensor_fit == "AUTO":
            if size_x >= size_y:
                return "HORIZONTAL"
            else:
                return "VERTICAL"
        return sensor_fit

    cam = cam or get_cams()[0]
    camd = cam.data
    if camd.type != "PERSP":
        raise ValueError("Non-perspective cameras not supported")
    scene = bpy.context.scene
    f_in_mm = camd.lens
    scale = scene.render.resolution_percentage / 100
    resolution_x_in_px = scale * scene.render.resolution_x
    resolution_y_in_px = scale * scene.render.resolution_y
    sensor_size_in_mm = get_sensor_size(
        camd.sensor_fit, camd.sensor_width, camd.sensor_height
    )
    sensor_fit = get_sensor_fit(
        camd.sensor_fit,
        scene.render.pixel_aspect_x * resolution_x_in_px,
        scene.render.pixel_aspect_y * resolution_y_in_px,
    )
    pixel_aspect_ratio = scene.render.pixel_aspect_y / scene.render.pixel_aspect_x
    if sensor_fit == "HORIZONTAL":
        view_fac_in_px = resolution_x_in_px
    else:
        view_fac_in_px = pixel_aspect_ratio * resolution_y_in_px
    pixel_size_mm_per_px = sensor_size_in_mm / f_in_mm / view_fac_in_px
    s_u = 1 / pixel_size_mm_per_px
    s_v = 1 / pixel_size_mm_per_px / pixel_aspect_ratio

    # Parameters of intrinsic calibration matrix K
    u_0 = resolution_x_in_px / 2 - camd.shift_x * view_fac_in_px
    v_0 = resolution_y_in_px / 2 + camd.shift_y * view_fac_in_px / pixel_aspect_ratio
    skew = 0  # only use rectangular pixels

    K = mathutils.Matrix(((s_u, skew, u_0), (0, s_v, v_0), (0, 0, 1)))
    return K


def set_cam_intrinsic(cam, K, hw=None):
    """
    Invert the function get_cam_intrinsic
    
    K =[[s_u, 0, u_0],
        [0, s_v, v_0],
        [0,   0,   1]]
    or
    K =[[f_x, 0, c_x],
        [0, f_y, c_y],
        [0,   0,   1]]
    """
    scene = bpy.context.scene
    camd = cam.data
    camd.sensor_fit = "HORIZONTAL"
    camd.sensor_width = 10
    scene.render.resolution_percentage = 100

    if hw is None:
        hw = scene.render.resolution_y, scene.render.resolution_x
    else:
        scene.render.resolution_y, scene.render.resolution_x = hw
    near = lambda x, y=0, eps=1e-5: abs(x - y) < eps
    assert near(K[0][1], 0)
    assert near(K[1][0], 0)

    h, w = hw
    s_u = K[0][0]  # f_x
    s_v = K[1][1]  # f_y
    u_0 = K[0][2]  # c_x
    v_0 = K[1][2]  # c_y
    scene.render.pixel_aspect_x = s_v / min(s_u, s_v)
    scene.render.pixel_aspect_y = s_u / min(s_u, s_v)
    pixel_aspect_ratio = s_u / s_v
    camd.lens = s_u * camd.sensor_width / w
    camd.shift_x = (w / 2 - u_0) / w
    camd.shift_y = (v_0 - h / 2) / w * pixel_aspect_ratio


if __name__ == "__main__":
    pass



#!/usr/bin/env python3

import boxx
from boxx import *

import bpy
import random

from .object_utils import activate_obj, load_obj


def texture_bsdf_dr(obj, shade_smooth_rate=0.9):
    """
    texture domain randomztion for object who has texture, like YCB's model.

    Parameters
    ----------
    obj : object
        object.
    shade_smooth_rate : float, optional
        shade_smooth_rate. The default is 0.9.

    Returns
    -------
    bsdf : Blender's node
        Principled BSDF.

    """
    if random.random() < shade_smooth_rate:
        with activate_obj(obj):
            bpy.ops.object.shade_smooth()

    material = obj.material_slots[0].material
    bsdf = material.node_tree.nodes["Principled BSDF"]
    bsdf.inputs["Metallic"].default_value = random.random()
    bsdf.inputs["Specular"].default_value = random.random()
    bsdf.inputs["Specular Tint"].default_value = random.random()
    bsdf.inputs["Roughness"].default_value = random.uniform(0.3, 1)
    bsdf.inputs["Anisotropic"].default_value = random.random()
    bsdf.inputs["Anisotropic Rotation"].default_value = random.random()
    bsdf.inputs["Sheen"].default_value = random.random()
    bsdf.inputs["Sheen Tint"].default_value = random.random()
    bsdf.inputs["Clearcoat"].default_value = random.random()
    bsdf.inputs["Clearcoat Roughness"].default_value = random.random()
    bsdf.inputs["Alpha"].default_value = random.uniform(0.9, 1)
    return bsdf


def load_distractor(distractor_path, target_size="fit_to_YCB"):
    """
    load_distractor, usually using shape net
    
    Notice: shape net's distractor should format by meshlabserver, than import to Blender

    Parameters
    ----------
    distractor_path : str
        distractor path.
    target_size : float, optional
        max side length's lenght. The default is "fit_to_YCB".
        will fit to YCB dataset's size

    Returns
    -------
    distractor.

    """
    if target_size == "fit_to_YCB":
        target_size = random.choice(
            [0.06, 0.1, 0.14, 0.17, 0.18, 0.2, 0.21, 0.23, 0.26, 0.3, 0.35]
        ) * random.uniform(0.8, 1.2)
    distractor = load_obj(distractor_path)
    with activate_obj(distractor):
        bpy.ops.object.origin_set(type="ORIGIN_GEOMETRY", center="MEDIAN")
    distractor.location = (0, 0, 0)
    xyh = [i - j for i, j in zip(distractor.bound_box[-2], distractor.bound_box[0])]
    max_distance = max(map(abs, xyh))
    scale = target_size / max_distance
    distractor.scale = (scale,) * 3
    return distractor


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Thu Jan  2 20:27:57 2020
"""

from boxx import *
from boxx import tree

import bpy
import time
import inspect
from functools import wraps


class StatuRecover:
    def __init__(self):
        self.obj_to_old_attr_value = []
        self.obj_to_old_bpy_prop_collection = {}

    def __enter__(self):
        return self

    def __exit__(self, typee, value, traceback):
        self.recover_statu()

    def recover_statu(self):
        for (obj, attr), value in self.obj_to_old_attr_value[
            ::-1
        ]:  # try avoid TypeError
            try:
                setattr(obj, attr, value)
            except TypeError as e:
                tree - (obj, attr, value)
                print(
                    "Maybe, This value is invalide for other attr, try change order of self.set_attr() in the code"
                )
                # raise e
        for (
            bpy_path,
            bpy_prop_collection,
        ) in self.obj_to_old_bpy_prop_collection.items():
            bpy_path.clear()
            list(map(bpy_path.append, bpy_prop_collection))

    def set_attrs(self, obj, attrs):
        for attr, value in attrs.items():
            self.set_attr(obj, attr, value)

    def set_attr(self, obj, attr, value):
        self.obj_to_old_attr_value.append([(obj, attr), getattr(obj, attr)])
        setattr(obj, attr, value)

    def replace_collection(self, bpy_path, bpy_prop_collection):
        self.obj_to_old_bpy_prop_collection[bpy_path] = bpy_path[:]
        bpy_path.clear()
        list(map(bpy_path.append, bpy_prop_collection))


class undo:
    """ Reverts all changes done to the blender project inside this block.
    Usage: 
        >>> with undo():
                # code
    Reference:
        https://github.com/DLR-RM/BlenderProc/blob/60e559065a4b09db2bd2425822c0ebde9c77609c/src/utility/Utility.py#L200
        
    Warning: undo() may invalid var who point to bpy object, which may cause lots problems.
    """

    def __init__(self, check_point_name=None):
        self.check_point_name = check_point_name

    def __enter__(self):
        check_point_name = self.check_point_name
        if check_point_name is None:
            check_point_name = (
                inspect.stack()[1].filename + " - " + inspect.stack()[1].function
            )
        self.check_point_name = check_point_name
        bpy.ops.ed.undo_push(message="before " + self.check_point_name)

    def __exit__(self, type, value, traceback):
        bpy.ops.ed.undo_push(message="after " + self.check_point_name)
        # The current state points to "after", now by calling undo we go back to "before"
        bpy.ops.ed.undo()

    def __call__(self, func):
        @wraps(func)
        def return_func(*l, **kv):
            with undo(str(time.time())):
                result = func(*l, **kv)
                return result

        return return_func


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Sun Jan  5 21:41:51 2020
"""

from boxx import *
from boxx import defaultdict


import bpy
import mathutils
import numpy as np

from .physic_utils import OLD_V0_KEY
from .camera_utils import get_cam_intrinsic


T_bcam2cv = np.array(((1, 0, 0, 0), (0, -1, 0, 0), (0, 0, -1, 0), (0, 0, 0, 1)))

# Returns camera rotation and translation matrices from Blender.
#
# There are 3 coordinate systems involved:
#    1. The World coordinates: "world"
#       - right-handed
#    2. The Blender camera coordinates: "bcam"
#       - x is horizontal
#       - y is up
#       - right-handed: negative z look-at direction
#    3. The desired computer vision camera coordinates: "cv"
#       - x is horizontal
#       - y is down (to align to the actual pixel coordinates
#         used in digital images)
#       - right-handed: positive z look-at direction
def get_4x4_world_to_cam_from_blender(camera):
    # bcam stands for blender camera
    R_bcam2cv = mathutils.Matrix(((1, 0, 0), (0, -1, 0), (0, 0, -1)))

    # Use matrix_world instead to account for all constraints
    location, rotation = camera.matrix_world.decompose()[0:2]
    R_world2bcam = rotation.to_matrix().transposed()

    # Convert camera location to translation vector used in coordinate changes
    # Use location from matrix_world to account for constraints:
    T_world2bcam = -1 * R_world2bcam @ location

    # Build the coordinate transform matrix from world to computer vision camera
    R_world2cv = R_bcam2cv @ R_world2bcam
    T_world2cv = R_bcam2cv @ T_world2bcam

    # put into 4x4 matrix
    world_to_cam = mathutils.Matrix(
        (
            R_world2cv[0][:] + (T_world2cv[0],),
            R_world2cv[1][:] + (T_world2cv[1],),
            R_world2cv[2][:] + (T_world2cv[2],),
            (0, 0, 0, 1),
        )
    )
    return world_to_cam


def get_K_world_to_cam(camera):
    K = get_cam_intrinsic(camera)
    world_to_cam = get_4x4_world_to_cam_from_blender(camera)
    return {
        "intrinsic_matrix": np.asarray(K, dtype=np.float32),
        "world_to_cam": np.asarray(world_to_cam, dtype=np.float32),
    }


def matrix_world_for_old_origin(obj):
    matrix_world = obj.matrix_world.copy()
    old_v0 = obj[OLD_V0_KEY]
    to_default_origin_vector = obj.data.vertices[0].co - mathutils.Vector(old_v0)
    matrix_world.translation = (
        matrix_world.translation + matrix_world.to_3x3() @ to_default_origin_vector
    )
    return matrix_world


def homo_coord(points_or_Rt):
    """
    homogeneous coordinates
    """
    if points_or_Rt.shape == (3, 4):
        return np.append(points_or_Rt, [[0, 0, 0, 1]], 0)
    elif points_or_Rt.shape[-1] in {2, 3}:
        return np.append(points_or_Rt, [[1]] * len(points_or_Rt), 1)
    return points_or_Rt


def set_matrix_world(obj, matrix_world):
    """
    Seems only set pose by obj.location and obj.rotation_euler 
    """
    if len(matrix_world) == 3:
        matrix_world = homo_coord(matrix_world)
    obj.location, quaternion, obj.scale = mathutils.Matrix(matrix_world).decompose()
    obj.rotation_euler = quaternion.to_euler()
    return obj


def get_pose_in_cam(obj, cam):
    pose_RT = (
        T_bcam2cv
        @ np.linalg.inv(np.array(cam.matrix_world))
        @ np.array(obj.matrix_world)
    )
    return pose_RT[:3]


def set_pose_in_cam(obj, pose_Rt, cam):
    if len(pose_Rt) == 3:
        pose_Rt = homo_coord(pose_Rt)
    matrix_world = npa(cam.matrix_world) @ T_bcam2cv.T @ pose_Rt
    set_matrix_world(obj, matrix_world)


def get_6d_pose(objs, inst=None, camera=None):
    def inst_id_to_area(inst_id):
        if inst is None:
            return -1
        return (inst == inst_id).sum()

    bpy.context.view_layer.update()
    if camera is None:
        camera = bpy.context.scene.camera
    meta = dict()
    meta.update(get_K_world_to_cam(camera))
    meta["cam_matrix_world"] = np.array(camera.matrix_world)
    for key in [
        "inst_ids",
        "areas",
        "visibles",
        "poses",
        "6ds",
        "bound_boxs",
        "mesh_names",
    ]:
        # default is []
        meta[key] = []
    for obj in objs:
        inst_id = obj.get("inst_id", -1)
        assert inst_id <= 100e4, f"inst_id '{inst_id}' should <= 100e4"
        area = inst_id_to_area(inst_id)
        if area != 0:
            meta["inst_ids"].append(inst_id)
            meta["areas"].append(area)
            meta["visibles"].append(area != 0)

            matrix_world = obj.matrix_world.copy()
            if OLD_V0_KEY in obj:
                matrix_world = matrix_world_for_old_origin(obj)

            pose = np.dot(meta["world_to_cam"], matrix_world)[:3]
            meta["poses"].append(pose[..., None])
            meta["6ds"].append(pose)
            bound_box = np.array([list(point) for point in obj.bound_box])
            meta["bound_boxs"].append(bound_box)
            meta["mesh_names"].append(obj.name)

    meta["poses"] = meta["poses"] and np.concatenate(meta["poses"], -1)
    return dict(meta)


if __name__ == "__main__":
    pass



import bpy
import boxx
import numpy as np


class StereoCam:
    def __init__(self, main_cam, pose_Rt, K=None, hw=None):
        """
        if K is None, K = mian_cam's K

        if pose_RT is int or float, mean x
        """
        import bpycv

        self.bpycv = bpycv

        self.main = main_cam
        self.cam = self.bpycv.duplicate(self.main)
        self.cam.name = self.main.name + "_stereo"
        pose_Rt = np.array(pose_Rt)
        if pose_Rt.size == 1:
            _pose_Rt = np.eye(4)[:3]
            _pose_Rt[0][-1] = pose_Rt
            pose_Rt = _pose_Rt
        self.pose_Rt = pose_Rt

        self.hw = (
            bpy.context.scene.render.resolution_y,
            bpy.context.scene.render.resolution_x if hw is None else hw,
        )
        if K is None:
            K = np.array(self.bpycv.get_cam_intrinsic(self.main))
        else:
            self.bpycv.set_cam_intrinsic(self.cam, K, hw)

        self.K = K
        self.inv_K = np.linalg.inv(K)
        self.T_main_to_stereo = np.linalg.inv(self.bpycv.homo_coord(pose_Rt))
        self.set_pose()

    def set_pose(self):
        self.bpycv.set_pose_in_cam(
            self.cam, self.bpycv.T_bcam2cv[:3, :3] @ self.pose_Rt, self.main
        )

    def render_image(self):
        self.set_pose()
        old_cam = bpy.context.scene.camera
        bpy.context.scene.camera = self.cam
        image = self.bpycv.render_image()
        bpy.context.scene.camera = old_cam
        return image

    def get_disparity(self, depth):
        """
        depth of main
        """
        hw = boxx.Vector(depth.shape)
        ys, xs = np.mgrid[: hw[0], : hw[1]]
        depth_mask = depth > 0
        xys_in_main = np.array([xs, ys, np.ones_like(xs)])[:, depth_mask].T
        xyzs_in_main = np.dot(xys_in_main * depth[depth_mask, None], np.linalg.inv(K).T)
        xyzs_in_stereo = np.dot(
            bpycv.homo_coord(xyzs_in_main), self.T_main_to_stereo.T
        )[:, :3]
        depth_in_stereo = xyzs_in_stereo[:, 2]
        xy_in_stereo_ = (
            np.dot(xyzs_in_stereo, self.K.T)[:, :2] / depth_in_stereo[..., None]
        )
        xy_in_stereo = xy_in_stereo_.round(0).astype(np.int32)

        in_hw_mask = (
            (xy_in_stereo >= 0).all(1)
            & (xy_in_stereo[:, 0] < hw.w)
            & (xy_in_stereo[:, 1] < hw.h)
        )
        # out_hw_mask = (xy_in_stereo<0).any(1) | (xy_in_stereo[:, 0] >= hw.w)| (xy_in_stereo[:,1] >= hw.h)

        _xys, inverse = np.unique(xy_in_stereo, axis=0, return_inverse=1)
        idxs = np.arange(depth_in_stereo.size)
        not_occlude_mask = np.zeros_like(depth_in_stereo, dtype=bool)
        for i in range(len(_xys)):
            same_point_mask = inverse == i
            idx = idxs[same_point_mask][np.argmin(depth_in_stereo[same_point_mask])]
            not_occlude_mask[idx] = True

        matched_mask = np.zeros_like(depth_mask)
        matched_mask[depth_mask] = in_hw_mask & not_occlude_mask

        disparity = xys_in_main[:, :2] / xy_in_stereo
        # TODO TBD

    def remove(self):
        self.bpycv.remove_obj(self.cam)


with boxx.impt("/home/dl/megvii/project/ai_asrs/stereo_depth/tools"):
    from calibration_stereo import CalibrationStereo


class CalibrationStereoBlender(CalibrationStereo):
    def __init__(self, Rt, K, hw, K2=None):
        import cv2

        K1 = K
        K2 = K if K2 is None else K2
        D1 = np.zeros((1, 5))
        D2 = np.zeros((1, 5))
        xy = tuple(hw[::-1])
        R, T = Rt[:3, :3], Rt[:3, 3:] * 1000
        R1, R2, P1, P2, Q, roi1, roi2 = cv2.stereoRectify(K1, D1, K2, D2, xy, R, T)
        self.data = dict(K1=K1, D1=D1, R1=R1, K2=K2, D2=D2, R2=R2, xy=xy, R=R, T=T)
        self._set_map()

    @classmethod
    def from_StereoCam(cls, stereo):
        cs = cls(
            stereo.pose_Rt,
            np.array(stereo.bpycv.get_cam_intrinsic(stereo.main)),
            stereo.hw,
            stereo.K,
        )
        return cs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Thu Jan 16 18:17:20 2020
"""
import boxx

import bpy


def remove_useless_data():
    """
    remove all data and release RAM
    """
    for block in bpy.data.meshes:
        if block.users == 0:
            bpy.data.meshes.remove(block)

    for block in bpy.data.materials:
        if block.users == 0:
            bpy.data.materials.remove(block)

    for block in bpy.data.textures:
        if block.users == 0:
            bpy.data.textures.remove(block)

    for block in bpy.data.images:
        if block.users == 0:
            bpy.data.images.remove(block)


def clear_all():
    [
        bpy.data.objects.remove(obj)
        for obj in bpy.data.objects
        if obj.type in ("MESH", "LIGHT", "CURVE")
    ]
    remove_useless_data()


def set_shading_mode(mode="SOLID", screens=[]):
    """
    Performs an action analogous to clicking on the display/shade button of
    the 3D view. Mode is one of "RENDERED", "MATERIAL", "SOLID", "WIREFRAME".
    The change is applied to the given collection of bpy.data.screens.
    If none is given, the function is applied to bpy.context.screen (the
    active screen) only. E.g. set all screens to rendered mode:
      set_shading_mode("RENDERED", bpy.data.screens)
    """
    screens = screens if screens else [bpy.context.screen]
    for s in screens:
        for spc in s.areas:
            if spc.type == "VIEW_3D":
                spc.spaces[0].shading.type = mode
                break  # we expect at most 1 VIEW_3D space


def add_stage(size=2, transparency=False):
    """
    add PASSIVE rigidbody cube for physic stage or depth background 

    Parameters
    ----------
    size : float, optional
        size of stage. The default is 2.
    transparency : bool, optional
        transparency for rgb but set limit for depth. The default is False.
    """
    import bpycv

    bpy.ops.mesh.primitive_cube_add(size=size, location=(0, 0, -size / 2))
    stage = bpy.context.active_object
    stage.name = "stage"
    with bpycv.activate_obj(stage):
        bpy.ops.rigidbody.object_add()
        stage.rigid_body.type = "PASSIVE"
        if transparency:
            stage.rigid_body.use_margin = True
            stage.rigid_body.collision_margin = 0.04
            stage.location.z -= stage.rigid_body.collision_margin

            material = bpy.data.materials.new("transparency_stage_bpycv")
            material.use_nodes = True
            material.node_tree.nodes.clear()
            with bpycv.activate_node_tree(material.node_tree):
                bpycv.Node("ShaderNodeOutputMaterial").Surface = bpycv.Node(
                    "ShaderNodeBsdfPrincipled", Alpha=0
                ).BSDF
            stage.data.materials.append(material)
    return stage


def add_environment_box(xyz=(2, 2, 2), thickness=0.2, transparency=False):
    import bpycv

    xyz = boxx.Vector(xyz)
    thickness += 1
    box = add_stage(size=1, transparency=transparency)
    box.rigid_body.collision_shape = "MESH"
    box.scale = xyz * thickness
    bpy.ops.mesh.primitive_cube_add(size=1, location=(0, 0, xyz.z / 2))
    cube = bpy.context.object
    box.location.z = cube.location.z - xyz.z * (thickness - 0.99)
    cube.scale = xyz
    with bpycv.activate_obj(box):
        bpy.ops.object.modifier_add(type="BOOLEAN")
        modifier = box.modifiers[-1]
        modifier.object = cube
        modifier.operation = "DIFFERENCE"
        if bpy.app.version >= (2, 91, 0):
            bpy.ops.object.modifier_apply(modifier=modifier.name)
        else:
            bpy.ops.object.modifier_apply(apply_as="DATA", modifier=modifier.name)
    bpycv.remove_obj(cube)
    return box


if __name__ == "__main__":
    pass
